/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "jftpgw.h"
#include "cmds.h"

/* these two functions reside in cmds.c */
int getuserdest(char*, struct clientinfo*);
int getpasswd(char*, struct clientinfo*);


/* the portcmd variable keeps track of the issued port */
extern char* portcmd;

int std_quit(char* args, struct conn_info_st* conn_info) {

	int ss = conn_info->clntinfo->serversocket;
	int cs = conn_info->clntinfo->clientsocket;
	
	/* Are we already connected to a server? */
	if (ss != -1) {
		char* response;
		say(ss, "QUIT\r\n");
		response = passall(ss, cs);
		conn_info->li->respcode = respcode(response);
		free(response);
	}
	else {
		/* Generate an own goodbye message if we are
		 * not yet connected */
		say(cs, "221 Goodbye...\r\n");
	}			
	return CMD_QUIT;
}


int std_user(char* args, struct conn_info_st* conn_info) {

	int ss = conn_info->clntinfo->serversocket;
	int cs = conn_info->clntinfo->clientsocket;
	
	/* USER specified though logged in already */
	if (ss != -1) {
		say(cs, "503 You are already logged in!\r\n");
		conn_info->li->respcode = 503;
		return CMD_HANDLED;
	}

	if (getuserdest(args, conn_info->clntinfo) == 0) {
		char* sendbuf = 0;
		size_t sendbufsize;
		const char* passreq =
			"331 Password required for %s.\r\n";
		conn_info->usergiven = 1;
		sendbufsize = strlen(passreq) 
				+ strlen(conn_info->lsforward->fwuser)
				+ 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		snprintf(sendbuf, sendbufsize, passreq,
			conn_info->lsforward->fwuser);
		say(cs, sendbuf);
		free(sendbuf);
		sendbuf = 0;
		conn_info->li->respcode = 331;
	}
	return CMD_HANDLED;
}

int std_pass(char* args, struct conn_info_st* conn_info) {
	
	int ss = conn_info->clntinfo->serversocket;
	int cs = conn_info->clntinfo->clientsocket;
	int usergiven = conn_info->usergiven;
	int ret;
	
	/* PASS specified though logged in already */
	if (ss != -1) {
		say(cs, "503 You are already logged in!\r\n");
		conn_info->li->respcode = 503;
		return CMD_HANDLED;
	}
	
	/* The client sent a PASS before a USER cmd!  */
	if (ss == -1 && !usergiven) {
		say(cs, "530 Please login with USER and PASS.\r\n");
		conn_info->li->respcode = 530;
		return CMD_HANDLED;
	}
	
	if (ss == -1 && usergiven) {
		if (!getpasswd(args, conn_info->clntinfo)) {
			ret = login(conn_info->clntinfo);
			if (!ret) {
				/* The login was successful */
				ss = conn_info->clntinfo->serversocket;
				free_loginst();
			}
			if (ret == -1) {
				/* perror reason */
				char* sendbuf = 0;
				const char* errmsg = 
					"500 Error connecting: %s\r\n";
				int err = errno;
				size_t sendbufsize;

				sendbufsize = strlen(errmsg)
					+ strlen(strerror(err))
					+ 1;
				sendbuf = (char*) malloc(sendbufsize);
				enough_mem(sendbuf);
				snprintf(sendbuf, sendbufsize,
						errmsg, strerror(err));
				say(cs, sendbuf);
				free(sendbuf);
				sendbuf = 0;
				conn_info->li->respcode = 500;
				return CMD_HANDLED;
			}
			if (ret < 0) {
				/* login failed - do not print an error
				 * message */
				conn_info->usergiven = 0;
				conn_info->clntinfo->serversocket = ss = -1;
			}
		}
	}
	return CMD_HANDLED;
}


int std_pasv(char* args, struct conn_info_st* conn_info) {
	char* answer = 0;
	int ret;

	if (conn_info->clntinfo->servermode == PASSIVE || 
	    conn_info->clntinfo->servermode == ASCLIENT) {
		ret = pasvserver(conn_info->clntinfo);
	} else {
		ret = activeserver(&answer, conn_info->clntinfo);
		conn_info->li->respcode = respcode(answer);
		free(answer);
	}
		
	ret |= pasvclient(conn_info->clntinfo);
	
	if (ret) {
		/* go on */
	}
	return CMD_HANDLED;
}


int std_port(char* args, struct conn_info_st* conn_info) {
	int ret = 0;
	size_t sendbufsize;
	int cs = conn_info->clntinfo->clientsocket;
	struct sockaddr_in sin;
	char *answer = 0;
	char *sendbuf = 0;

	if (portcommandcheck(args, &sin, conn_info->clntinfo) < 0) {
		/* an error was already reported */
		return -1;
	}
	portcmd = strdup(args);
	enough_mem(portcmd);

	if (conn_info->clntinfo->servermode == ACTIVE ||
	    conn_info->clntinfo->servermode == ASCLIENT) {
		ret = activeserver(&answer, conn_info->clntinfo);
		if (ret < 0) {
			free(answer);
			return CMD_HANDLED;
		}
		/* Repeat the answer of the server */
		sendbufsize = strlen(answer) + 3;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		snprintf(sendbuf, sendbufsize, "%s\r\n", answer);
		say(cs, sendbuf);
		conn_info->li->respcode = respcode(sendbuf);
		free(sendbuf);
		free(answer);
	} else {
		ret |= pasvserver(conn_info->clntinfo);
		if (!ret) {
			conn_info->li->respcode = 200;
			say(cs, "200 PORT command successful.\r\n");
		}

	}
	return CMD_HANDLED;
}


int std_stor(char* args, struct conn_info_st* conn_info) {
	/* chop of the "STOR " prefix */
	char* space = strchr(args, ' ');
	if (space) {
		conn_info->li->name = space + 1;
	} else {
		conn_info->li->name = args;
	}

	conn_info->li->direction = 'i';
	conn_info->clntinfo->mode = STOR;
	return CMD_PASS;
}

int std_retr(char* args, struct conn_info_st* conn_info) {
	/* chop of the "RETR " prefix */
	char* space = strchr(args, ' ');
	if (space) {
		conn_info->li->name = space + 1;
	} else {
		conn_info->li->name = args;
	}
	conn_info->li->direction = 'o';
	return CMD_PASS;
}

/* a simple function that determines the transfer mode (ascii or image) */

int std_type(char* args, struct conn_info_st* conn_info) {
	char* space = strrchr(args, ' ');
	if (space) {
		if (*(space + 1) == 'A' || *(space + 1) ==  'a') {
			conn_info->li->type = 'a';
		} else {
			conn_info->li->type = 'b';
		}
	}
	return CMD_PASS;
}

